/*	$Id: lmin.c,v 3.8 1998/05/23 23:35:18 kas Exp $	*/

/* Local minimization. Calls NAG's e04dgf_() or Minpack's dvmlm_() */

/*
 * This file was written by Petr Mejzlk and modified for the
 * new gradient handling by Jan "Yenya" Kasprzak.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "proto.h"
#include "ff.h"
#include "pval.h"

int lmin_verbose = 0;
/*** Headers for local minimization procedures written in FORTRAN ***/

#ifdef FLONGINT
typedef long int integer; /* Fortran integer */
typedef long int ftnlen;  /* string length in Fortran */
#else
typedef int integer;
typedef int ftnlen; 
#endif

#ifdef NAG
/* 1. Conjugate gradient with limited memory BFGS for updates from
   NAG Fortran Library Mark 16. */

void e04dgf_(integer *n, 
  void (*fungrd)(integer *mode, integer *n, double x[], double *objf, 
                 double grd[], integer *nstate, integer iuser[], double user[]),
  integer *iter, double *objf, double objgrd[], double x[], integer iwork[], 
  double work[], integer iuser[], double user[], integer *ifail);

void e04dkf_(char *string, ftnlen length_1);

static void field(integer *mode, integer *nvars, double vars[], double *objf, 
  double grd[], integer *nstate, integer iuser[], double user[]);

static potential_value *my_pvp;

#else
/* 2. Limited memory BFGS from MINPACK-2 (directory vmlm). */

void dvmlm_(integer *n, double x[], double *objf, double objgrd[], 
  double *frtol, double *fatol, double *fmin, char *task, 
  integer *m, double s[], double y[], double rho[], 
  integer isave[5], double dsave[24],
  double wa1[], double wa2[], ftnlen task_l);

static void field(potential_value *pvp, double *objf, double grd[]);
#endif

#define max(A, B) ((A) > (B) ? (A) : (B))
#define DET(a11,a12,a21,a22) ((a11)*(a22)-(a21)*(a12))

static int numgrad;
static int iters;

REAL minimize_l(pvp, precision, numeric)
potential_value *pvp;
double precision; /* absolute accuracy of the result */
int numeric;
{
	integer dim;

#ifdef NAG
        double *w, *scratch, f;
        integer *iw, liw, lw, niter, ifail, *iscratch;
        double maxstep;
	char str[50];
#else /* Minpack */
	double *s, *y, *rho, *wa1, *wa2, f, dsave[24];
	double frtol=1.0e-12, fatol=1.0e-10, fmin=-1.0e30;
	char task[60];
	integer m=5, isave[5];
	ftnlen task_l;
#endif

	numgrad = numeric;
	iters = 0;

	if (sizeof(REAL) != sizeof(double)) {
	        perror("lmin.c: REAL must be double");
	        exit(1);
	}
        dim = pvp->ffp->natoms*3; /* number of variables */

#ifdef NAG
        liw = dim+2; /* length of integer workspace */
        lw = 13*dim; /* length of floating-point workspace (double) */

        /* allocate workspace required by e04dgf_ : */
        if (((iw = (integer *) malloc(sizeof(int)*liw)) == NULL)      ||
            (( w = (double *) malloc(sizeof(double)*lw)) == NULL)     ||
            ((scratch = (double *) malloc(sizeof(double)*4*dim)) == NULL) ||
            ((iscratch = (integer *) malloc(sizeof(int)*dim)) == NULL) )

                return pvp->value;

        my_pvp=pvp;
        ifail = -1; /* soft return in the case of failure */

        e04dkf_("Nolist", 6);
        e04dkf_("Print Level =  0", 16);
        e04dkf_("Verify Level = -1", 17);
        maxstep = 0.75;
        sprintf(str,"Maximum Step = %5.3f", maxstep);
        e04dkf_(str, 20);
        e04dkf_("Linesearch Tolerance = 0.9", 26);
        sprintf(str,"Optimality Tolerance = %8.3e", precision);
        e04dkf_(str, 31);

        e04dgf_(&dim, field, &niter, &f, pvp->dx, pvp->x, iw, w, 
                iscratch, scratch, &ifail);

        pvp->value = f;

        free(iw), free(w);
        free(iscratch), free(scratch);

#else /* Minpack */
	fatol = precision;

	/* allocate workspace required by dvmlm_ : */
	if (((s = (double *) malloc(sizeof(double)*dim*m)) == NULL)   ||
	    ((y = (double *) malloc(sizeof(double)*dim*m)) == NULL)   ||
	    ((rho = (double *) malloc(sizeof(double)*m)) == NULL)     ||
	    ((wa1 = (double *) malloc(sizeof(double)*dim)) == NULL)   ||
	    ((wa2 = (double *) malloc(sizeof(double)*m)) == NULL) )

		return pvp->value;

	/* Initial function & gradient evaluation: */
	
	field(pvp, &f, pvp->dx);
	
	/* Initial call to dvmlm_: */

	strcpy(task, "START"); task_l = 5;
	do {
		/* Minimization step: */
		dvmlm_(&dim, pvp->x, &f, pvp->dx, &frtol, &fatol, &fmin,
		       task, &m, s, y, rho,
        	       isave, dsave, wa1, wa2, task_l);

		if (strncmp(task, "FG", 2) == 0) 
			field(pvp, &f, pvp->dx);
		else if (strncmp(task, "NEWX", 4) == 0) 
			; /* printf("%f\n", f); */
		else break;

	} while (1);
	
        pvp->value = f;

	free(s); free(y); free(rho); free(wa1); free(wa2);
#endif

	if (lmin_verbose >= 3)
		printf("%d iterations\n", iters);
	return compute_potential(pvp);
}

#ifdef NAG
static void field(integer *mode, integer *nvars, double vars[], double *objf, 
		  double grd[], integer *nstate, integer iscratch[], 
		  double scratch[])
#else /* Minpack */
static void field(potential_value *pvp, double *objf, double grd[])
#endif
{
	int dim;
  #ifdef PRINT_DEBUG
        static int fcount=0, gcount=0;
  #endif

#ifdef NAG
        static REAL *xyz0;
#else /* Minpack */
	double *vars;
	potential_value *my_pvp;
#endif

#ifdef NAG
        /* Save the address of the vector of variables in my_pvp: */
        xyz0 = my_pvp->x;
        /* Substitute the vector comming from the optimization procedure
           for the original vector of variables in my_pvp: */
        my_pvp->x = vars;

	dim = *nvars;

        if (vars != my_pvp->x) {
                printf ("Invalid call of field() -- need to save pvp->x");
                exit(1);
        }
        if (*nvars != 3*my_pvp->ffp->natoms) {
                printf ("Weird: In field(): Got array of %d numbers "
                        "(expecting %d from my_pvp->ffp->natoms)\n",
                        dim, 3*my_pvp->ffp->natoms);
                exit(1);
        }

        if (*mode == 0 || *mode == 2) {

#else /* Minpack */

        dim = 3*pvp->ffp->natoms;
	vars = pvp->x;
	my_pvp = pvp;
#endif

		/* Compute potential: */

  	#ifdef PRINT_DEBUG
		printf("field %d: starting\n",++fcount);
  	#endif
		*objf = compute_potential(my_pvp);
		if (lmin_verbose >= 4) 
			printf("iteration %7d, value %15.3f\n", iters, *objf);
	#ifdef PRINT_DEBUG
                printf("field %d: %10.5f\n",fcount,*objf);
		for(i=0;i<dim;i++) printf("%7.3f",vars[i]); printf("\n");
	#endif

#ifdef NAG
        }
        if (*mode == 2) { /* compute the gradient */
#endif
		/* Compute the gradient: */
		if (grd != pvp->dx) {
			printf("field: got different gradient vector!\n");
			exit(0);
		}
		if (numgrad)
			compute_grad_num(pvp);
		else
			compute_gradient(pvp);
#ifdef NAG
        }
        /* Restore the old vector of variables: */
        my_pvp->x = xyz0;
#endif
	iters++;
}
